Overview
========
This example demonstrates how to develop a remote IO device based on Slave Stack Code(SSC).

This remote IO device only supports SM-Synchron mode for DC operation, and implements one Input channel and one Output channel:
	Output channel: connected with the D6 LED and used to control this LED.
	Input channel: used to read the D6 LED current status. 

Note that the demo CAN NOT be built directly. Please see the "Prepare the Demo" section for more details.


SDK version
===========
- Version: 2.16.000

Toolchain supported
===================
- IAR embedded Workbench  9.60.1
- Keil MDK  5.39.0
- GCC ARM Embedded  13.2.1

Hardware requirements
=====================
- MIMXRT1180-EVK RevC
- RJ45 Network cable
- Mini/micro USB cable
- Personal Computer on which the TwinCat3 has been installed

Board settings
==============
For multiple boards cascades, R142 register needs to be removed.

Prepare the Demo
================
1. Generate the SSC source code
	- Download Slave Stack Code Tool(SSC Tool) from BECKHOFF official website and install it
	- Open Slave Stack Code Tool
	- Click 'File' -> 'New' -> 'Import' to import the Congfig file
		Select digital_io.xml under path '.\boards\evkmimxrt1180\ecat_examples\digital_io\<core>\SSC'.
	- Click 'Tool' -> 'Application' -> 'Import' to import Application information
		Select digital_io.xlsx under path '.\boards\evkmimxrt1180\ecat_examples\digital_io\<core>\SSC'.
	- Click 'Project' -> 'Create new Slave Files'
		'Source Folder'(default): '.\boards\evkmimxrt1180\ecat_examples\digital_io\<core>\SSC\Src'.
		Click 'Start'.
	- Remove digital_io.c under Src folder
	
2. Connect EtherCAT Port0 with TwinCAT3
	RJ45 interface labeled as J28 is the EtherCAT Port0 and RJ45 interface labeled as J32 is the EtherCAT Port1.

3. Connect a USB cable between the host PC and the OpenSDA USB port on the target board

4. Open a serial terminal with the following settings
	- 115200 baud rate
	- 8 data bits
 	- No parity
	- One stop bit
	- No flow control

5. Download the program to the target board

6. Either press the reset button on your board or launch the debugger in your IDE to begin running the demo

7. Update ESI file to EEPROM
	- Copy ESI file 'ECAT-IO.xml' generated by SSC TOOL to <TwinCAT_installation_folder>/<Version>/Config/io/EtherCAT/
	- Restart TwinCAT3 IDE
	- Scan device and Update EEPROM

8. Scan and load this device and create a PLC project in TwinCat3
	The PLC Main code is as follows:
		PROGRAM MAIN
		VAR 
			Delay : TON;
			LedIn AT %I* : BOOL; 
			LedOut AT %Q* : BOOL;
		END_VAR
		
		Delay(IN := TRUE, PT := T#1S);
		IF Delay.Q THEN
			LedOut := NOT LedIn;
			Delay(IN := FALSE);
		END_IF;
		
9. Activate configuration and start PLC Main function

Running the demo
================
When the example is running, the serial port will output:

       Start the SSC digital_io example...

After the PLC program started, the D6 LED starts to blink with 0.5Hz frequency.

NOTE: refer to the User Guide for the details of how to setup the PLC project on TwinCAT3.

